# FaunaDB Data Manager

The FaunaDB Data Manager (FDM) is a command-line application that
simplifies moving small and large amounts of data in and out of a
FaunaDB database.

The FDM can acquire data from a variety of different sources to assist
with a variety of import and export tasks. First is its ability to
acquire data from a single file, a set of files, or directory full of
files. It examines each file's contents and auto-detects if the file is
of type JSON or CSV; any other file type is ignored. The FDM uses a
file's base name as the collection name, and if the collection does not
exist it is automatically created. Some of the common use cases of this
type of input include the initial loading of a database, and incremental
loading when data is derived from a secondary system.

The FDM can also import data from a JDBC server. This allows the FDM to
connect directly to a third-party database and extract tables into a
FaunaDB database without any intermediate files. A common use case is
the transition from a legacy database system, such as MySQL or
PostgreSQL, to a FaunaDB database.

Lastly, the FDM can utilize an existing FaunaDB database as a data
source.  Some of the use cases include taking a snapshot export in JSON
of the data and schema to a local directory. Another use case is a
snapshot copy of an existing FaunaDB database to a new FaunaDB database.
Importantly, both schema (Collections, Indexes, Role, Functions) and
data (Documents) are copied in this process!


## Usage

To use the FaunaDB Data Manager, please download and run the
pre-compiled JAR file here:

https://fauna-repo.s3.amazonaws.com/fdm/fdm.zip


## Supported Input File Formats

The loader automatically determines the type of file by reading the
first line of the load file.

| File Type       | Description |
| ----------------| ------------- |
| JSON-Line       | Each line contains a single JSON document. |
| JSON-Array | The first element in the file is an array and the second is a document. Each document in the array is inserted as a seperate document into the FaunaDB database. |
| CSV with Headers  | Comma-Separated-Value where the headers (i.e. the first line in the file) are used as field names.|


## Requirements

Java 11 or greater


## Command Line Arguments

Command line arguments take precedence over any option set in the
`fdm.props` file. If an option is not passed on the command line, only
then is the property in the `fdm.props` file used. When a property is
not found, an application default is used.

```
Usage: FDM   1.14
=================
 -c, --config <path>  Specify the path to the configuration file
 --demo <key>         Create a demo database
 -d, --dest <arg>     The destination options for the data:
                         path=<export directory>
                         key=<secret>
                            uri=<prot://host:port>]
                            collection=[exists | replace ] default=exists
                            index=[exists | replace | ignore] default=exists
                            function=[exists | replace | ignore] default=exists
                            role=[exists | replace | ignore] default=exists
                            data=[copy | ignore] default=copy
 -f, --format <arg>   Specify the field transformation options:
                         <field-name>[->new-name]:<field-type>[(date_format)]
 -h, --help           Print this help text
 -s, --source <arg>   The source option(s) for the data (indented=related):
                         path=<load file or directory>
                         key=<secret>
                            [uri=<prot://host:port>]
                            [pit=<timestamp>]
                         aws=<secret>
                            id=<id>
                            region=<region>
                            bucket=//bucket/folder
                         jdbc=<path>
                            driver=<name>
                            url=<jdbc_url>
                            user=<user>
                            password=<pass>
                            [database=<dbs>]
                            [table=<table>]
 --tsize <size>       Specify the number of documents per transaction
                         Default: 100
 -v, --verbose        Enable verbose output
 -V, --version        Print the tool's version


Examples
===========================================================================

- Copy a database at a point in time:
  fdm -source key=<source_secret> pit=12/1/2018 -dest key=<dest_secret>

- Export a database to the directory '/work/export':
  fdm -source key=<source_secret> pit=12/1/2018 -dest path=/work/export

- Import all JSON and CSV files from the directory '/work/import':
  fdm -source path=/work/import -dest key=<dest_secret>

- Test import of all JSON and CSV files from AWS bucket and folder:
  fdm -source aws=${AWS_SECRET_ACCESS_KEY} id=${AWS_ACCESS_KEY_ID} \
      region=${AWS_DEFAULT_REGION} bucket=//bucket/load -dest dryrun
```


## The `fdm.props` file

The `fdm.props` file is an optional file where commonly-used command line
options are placed. If you are loading several files, you can place the
host name and key in the `fdm.props` file so that these two parameters
do not have to be provided on each command line.

    load
    Description:    A filename, set of filenames or directory where the
                    files to load are to be located.  The type of file
                    (JSON or CSV) will be automatically determined by
                    examine the first line of the file.
    Default:        data.json
    Example:        load=customer.json
    Example:        load=customer.json,names.json,employee.json
    Example:        load=/work/my_directory

    policy.collection:
    Description:    The Collection Schema policy [exists | replace]
    Default:        exists

    policy.index:
    Description:    The Index Schema policy [exists | replace | ignore]
    Default:        exists

    policy.function:
    Description:    The Function Schema policy [exists | replace | ignore]
    Default:        exists

    policy.role:
    Description:    The Role Schema policy [exists | replace | ignore]
    Default:        exists

    policy.data:
    Description:    The Data policy [copy | ignore]
    Default:        copy

    transactionSize
    Description:    The number of records to bundle into a single
                    transaction.
    Default:        100
    Example:        transactionSize=50

    format
    Description:    This allows you to control how an individual
                    field/value should be converted. You can use the
                    format to change the name of a field or cast the
                    data to a new type.
    Default:        None
    Example:        format=power_usage->used_power,code:long

    tsFormat
    Description:    Allow a global timestamp format to be set.
    Default:        yyyy-MM-dd'T'HH:mm:ss'Z'

    dateFormat
    Description:    Allow a global date format to be set.
    Default:        MM-dd-yyyy


## Data Transformation Details

The format statement allows you to rename fields, ignore fields or cast
values to new data types. The general layout of the format:

    <field-name>[->new-name]:<field-type>[(date_format)],...


| Element      | Argument    | Description  |
| -------------| ----------- | ------------ |
| field-name   | required    | The name of the field seen in the source data file. |
| new-name     | optional    | The new name of this field in the database.  This provides provides the ability to rename fields. |
| field-type   | optional    | Tells the loader what data type to load load the field as. See `Support types` below |
| (date_format) | optional    | An optional argument to field-type. The `date` and `ts`/`timestamp` types can accept a date format string to enhance the ability to convert to their target data type.|


### Supported Field Types:

The field type names are case sensitive and allow you to check that the
data is loaded into the requested type. If the conversion can not be
made, then a default value is placed is used and no error is reported.

    boolean:    Converts the field value to a boolean.
    
                Booleans map
                naturally; integer numbers other than 0 map to true,
                0 maps to false, and the strings 'true' and 'false' map
                to their corresponding boolean values.
    
    long:       Converts the field value to a long integer.
    
                Numbers are coerced using default Java rules; booleans
                convert to 0 (false) and 1 (true), and strings are
                parsed using default Java language integer parsing
                rules.
    
    double:     Converts the field value to a double-precision floating
                point number.
    
                Numbers are coerced using default Java rules; booleans
                convert to 0.0 (false) and 1.0 (true), and Strings are
                parsed using default Java language integer parsing
                rules.
    
    number:     Converts the field value to a long, if possible,
                otherwise converts the field value to a double.
    
    string:     Converts the field value to a string, if possible,
                otherwise an empty string.
    
    date:       Converts the field value to a date, if possible.
    
    ts:         Converts the field value to a timestamp, if possible.
    
    timestamp:  An alias for ts.
    
    ignore:     Do not place this field and its value into the
                destination document.
    
    ID:         Use the field value as the id in the destination
                document's reference.
    
                If this field name is found in the source document and
                this field is a long value, then this field is used as
                the document's reference's id. If this value is not set
                then the document's reference id is automatically
                generated by the database server.
    
    REF:        An alias for ID.
    
    TS:         Use the field value as the destination document's
                creation time.
    
                If this field name is found in the source
                document, then its value is used as the destination
                document creation timestamp, and the field is removed
                from the destination document. If this field is not
                found, or an incorrect format is found, then the current
                time is used.
    
                The config for TS is the format of the datetime in the
                source document (yyyy-MM-dd'T'HH:mm:ss.SSSZ). If the
                timestamp in the source document cannot be converted,
                then the current date and time is used for the
                destination document's creation time.  In addition, a
                warning message is displayed.


### Renaming A Field

To rename a field, use the format parameter and provide both the
current field name and the new field name, separated by an arrow (->):
`current_field_name->new_field_name`.

For example, when a source document contains the field called
`power_usage` but you want the destination document's field to have the
name `used_power`, you would specify the following:

    power_usage->used_power


## Policies

The default behavior for all schema objects is to leave the current
schema object if it already exists. The proposed solution would allow
the user to change the default policy for each type of schema object
(Collections, Indexes, Functions, Roles, and Data).


### Definition of Schema Policy Actions

| Policy| Description|
|-------|------------|
|exist  | if exists, use existing object |
|replace | if exists drop and replace with the new |
|ignore | do not move schema objects to destination |



### Definition of Data Policy Actions

| Policy| Description|
|-------|------------|
|copy | copy the data from source to destination |
|ignore | do not move schema objects to destination |


## Examples

Command line examples of running the FDM.


### Loading a single file

```
$ fdm -source path=/tmp/load/data100_basic.json -dest key=secret host=127.0.0.1
2020-04-09.15.51.01
===================

FDM   1.14
2020-04-09 22:51
Loaded property file path: /private/tmp/tmp/fdm-1.14/fdm.props

Field Formatting
================
<<None>>

Config File
===========
   load:                      data.json
   port:                      8443
   host:                      127.0.0.1
   policy.data:               COPY

Connecting to: http://127.0.0.1:8443
Connection was successful.
Starting Load at 2020-04-09 15:51:03.
File data100_basic.json -> data100_basic documents 100 size 11 KB  execution time:    0.015 sec  0.737 MB/sec  6666.7 docs/sec
FDM finished at 2020-04-09 15:51:04.

```

### Load a directory full of files


```
fdm -source path=/tmp/load  -dest key=secret host=127.0.0.1
2020-04-09.15.51.05
===================

FDM   1.14
2020-04-09 22:51
Loaded property file path: /private/tmp/tmp/fdm-1.14/fdm.props

Field Formatting
================
<<None>>

Config File
===========
   load:                      data.json
   port:                      8443
   host:                      127.0.0.1
   policy.data:               COPY

Connecting to: http://127.0.0.1:8443
Connection was successful.
Starting Load at 2020-04-09 15:51:07.
File data10_basic.csv -> data10_basic documents 10 size 0 KB  execution time:    0.001 sec  0.207 MB/sec  10000.0 docs/sec
File data100_basic.json -> data100_basic documents 100 size 11 KB  execution time:    0.015 sec  0.737 MB/sec  6666.7 docs/sec
File jay.csv -> jay documents 5 size 2 KB  execution time:    0.003 sec  0.777 MB/sec  1666.7 docs/sec
File data10_id_ts.json -> data10_id_ts documents 10 size 1 KB  execution time:    0.003 sec  0.637 MB/sec  3333.3 docs/sec
File data1000_array.json -> data1000_array documents 1000 size 147 KB  execution time:    0.015 sec  9.584 MB/sec  66666.7 docs/sec
File data100_array.json -> data100_array documents 100 size 11 KB  execution time:    0.002 sec  5.578 MB/sec  50000.0 docs/sec
File data1000_basic.json -> data1000_basic documents 1000 size 146 KB  execution time:    1.128 sec  0.127 MB/sec  886.5 docs/sec
File data10Types.csv -> data10Types documents 10 size 0 KB  execution time:    0.002 sec  0.347 MB/sec  5000.0 docs/sec
File data10.csv -> data10 documents 10 size 0 KB  execution time:    0.001 sec  0.292 MB/sec  10000.0 docs/sec
FDM finished at 2020-04-09 15:51:10.
```


### Starting the product  with a mysql JDBC driver

```
fdm -source jdbc=/work/JDBC/mysql-connector-java-8.0.18.jar  driver=com.mysql.cj.jdbc.Driver url=jdbc:mysql://localhost:3306/demo user=bob password=pass database=demo  -dest dryrun
```

MySQL settings

| Setting          | Value  |
| ----------------| ------------- |
| Driver  | com.mysql.cj.jdbc.Driver |
| URL  | jdbc:mysql://localhost:3306/DatabaseName|
| user  | connecting user name|
| password  | connecting user password|


## Authors

* **John Miller** - [GitHub](https://github.com/jfmiii)
